@@ -1179,11 +1179,11 @@ char *file, *def;
 		if (haste) keywaitfunc = readstatus;
 #endif
 #ifdef	_NOEDITMODE
-		ch = Xgetkey(SIGALRM);
+		ch = Xgetkey(SIGALRM, 0);
 #else
-		Xgetkey(-1);
-		ch = Xgetkey(SIGALRM);
-		Xgetkey(-1);
+		Xgetkey(-1, 0);
+		ch = Xgetkey(SIGALRM, 0);
+		Xgetkey(-1, 0);
 #endif
 #ifndef	_NOPRECEDE
 		keywaitfunc = NULL;
