@@ -17,13 +17,14 @@
 
 #if	!MSDOS
 #include <sys/file.h>
+#include <sys/param.h>
 #endif
 
 #ifndef	_NODOSDRIVE
 extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
 #endif
 
-extern int columns;
+extern int curcolumns;
 extern int mark;
 extern off_t marksize;
 extern off_t blocksize;
@@ -190,15 +191,15 @@ bindtable bindlist[MAXBINDTABLE] = {
 	{K_F(19),	VIEW_FILE,	255},
 #ifndef	_NOARCHIVE
 	{K_F(20),	PACK_FILE,	255},
-	{CR,		LAUNCH_FILE,	IN_DIR},
+	{K_CR,		LAUNCH_FILE,	IN_DIR},
 #else
-	{CR,		VIEW_FILE,	IN_DIR},
+	{K_CR,		VIEW_FILE,	IN_DIR},
 #endif
 	{K_BS,		OUT_DIR,	255},
 	{K_DC,		PUSH_FILE,	255},
 	{K_IC,		POP_FILE,	255},
 	{'\t',		MARK_FILE,	255},
-	{ESC,		QUIT_SYSTEM,	255},
+	{K_ESC,		QUIT_SYSTEM,	255},
 
 	{'<',		CUR_TOP,	255},
 	{'>',		CUR_BOTTOM,	255},
@@ -423,7 +424,7 @@ char *arg;
 static int one_column(arg)
 char *arg;
 {
-	columns = 1;
+	curcolumns = 1;
 	return(2);
 }
 
@@ -431,7 +432,7 @@ char *arg;
 static int two_columns(arg)
 char *arg;
 {
-	columns = 2;
+	curcolumns = 2;
 	return(2);
 }
 
@@ -439,7 +440,7 @@ char *arg;
 static int three_columns(arg)
 char *arg;
 {
-	columns = 3;
+	curcolumns = 3;
 	return(2);
 }
 
@@ -447,7 +448,7 @@ char *arg;
 static int five_columns(arg)
 char *arg;
 {
-	columns = 5;
+	curcolumns = 5;
 	return(2);
 }
 
@@ -585,12 +586,11 @@ char *arg;
 static int in_dir(arg)
 char *arg;
 {
-	if (
 #ifndef	_NOARCHIVE
-	!archivefile &&
+	if (archivefile);
+	else
 #endif
-	!strcmp(filelist[filepos].name, "."))
-		return(warning_bell(arg));
+	if (!strcmp(filelist[filepos].name, ".")) return(warning_bell(arg));
 	return(5);
 }
 
@@ -724,7 +724,8 @@ char *file;
 	tflush();
 	buf = NEXT_K;
 	i = strlen3(file);
-	if (i + strlen(buf) > n_lastcolumn) i = n_lastcolumn - strlen(buf);
+	if (i + strlen(buf) > n_lastcolumn)
+		i = n_lastcolumn - (int)strlen(buf);
 	prompt = malloc2(i * KANAWID + strlen(buf) + 1);
 	strncpy3(prompt, file, &i, 0);
 	strcat(prompt, buf);
@@ -767,7 +768,7 @@ char *env, *arg;
 static int NEAR execshell(VOID_A)
 {
 	char *sh;
-	int ret;
+	int n, ret;
 
 	sh = getenv2("FD_SHELL");
 #if	MSDOS
@@ -801,8 +802,8 @@ static int NEAR execshell(VOID_A)
 	}
 #endif
 
-	sigvecreset();
-	putterms(t_end);
+	n = sigvecset(0);
+	if (n) putterms(t_end);
 	stdiomode();
 	kanjifputs(SHEXT_K, stderr);
 	fputc('\n', stderr);
@@ -812,9 +813,9 @@ static int NEAR execshell(VOID_A)
 	if (sh) ret = dosystem(sh);
 	else ret = shell_loop(1);
 #endif
-	putterms(t_init);
+	if (n) putterms(t_init);
 	ttyiomode();
-	sigvecset();
+	sigvecset(n);
 
 	return(ret);
 }
@@ -932,7 +933,7 @@ char *arg;
 			i = 5;
 			tmp1 = val[0];
 		}
-		if (selectstr(&tmp1, i, 0, str, val) == ESC) return(1);
+		if (selectstr(&tmp1, i, 0, str, val) == K_ESC) return(1);
 	}
 
 	if (!tmp1) {
@@ -944,7 +945,7 @@ char *arg;
 		str[1] = ODEC_K;
 		val[0] = 0;
 		val[1] = 8;
-		if (selectstr(&tmp2, 2, 56, str, val) == ESC)
+		if (selectstr(&tmp2, 2, 56, str, val) == K_ESC)
 			return(1);
 		sorton = tmp1 + tmp2;
 		dupl = (int *)malloc2(maxfile * sizeof(int));
@@ -1078,6 +1079,9 @@ char *arg;
 static int rename_file(arg)
 char *arg;
 {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char path[MAXPATHLEN];
+#endif
 	char *file;
 
 	if (isdotdir(filelist[filepos].name)) return(warning_bell(arg));
@@ -1103,7 +1107,12 @@ char *arg;
 		free(file);
 	}
 
-	if (Xrename(filelist[filepos].name, file) < 0) {
+#if	MSDOS || defined (_NODOSDRIVE)
+	if (Xrename(filelist[filepos].name, file) < 0)
+#else
+	if (Xrename(nodospath(path, filelist[filepos].name), file) < 0)
+#endif
+	{
 		warning(-1, file);
 		free(file);
 		return(1);
@@ -1119,8 +1128,6 @@ char *arg;
 static int delete_file(arg)
 char *arg;
 {
-	int ret;
-
 	removepolicy = 0;
 	if (mark > 0) {
 		if (!yesno(DELMK_K)) return(1);
@@ -1129,9 +1136,7 @@ char *arg;
 	else if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	else {
 		if (!yesno(DELFL_K, filelist[filepos].name)) return(1);
-		if ((ret = rmvfile(filelist[filepos].name)) < 0)
-			warning(-1, filelist[filepos].name);
-		if (!ret) filepos++;
+		filepos = applyfile(rmvfile, NULL);
 	}
 	if (filepos >= maxfile && (filepos -= 2) < 0) filepos = 0;
 	return(4);
@@ -1148,9 +1153,15 @@ char *arg;
 	if (islink(&(filelist[filepos]))) {
 		int ret;
 
-		if ((ret = rmvfile(filelist[filepos].name)) < 0)
-			warning(-1, filelist[filepos].name);
-		if (!ret) filepos++;
+# ifdef	_NODOSDRIVE
+		ret = rmvfile(filelist[filepos].name);
+# else
+		char path[MAXPATHLEN];
+
+		ret = rmvfile(nodospath(path, filelist[filepos].name));
+# endif
+		if (ret < 0) warning(-1, filelist[filepos].name);
+		else if (!ret) filepos++;
 	}
 	else
 #endif
@@ -1369,7 +1380,16 @@ char *arg;
 
 	if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	if (archivefile) ret = unpack(archivefile, NULL, arg, 0, 0);
-	else ret = unpack(filelist[filepos].name, NULL, arg, 0, 1);
+	else {
+#if	MSDOS || defined (_NODOSDRIVE)
+		ret = unpack(filelist[filepos].name, NULL, arg, 0, 1);
+#else
+		char path[MAXPATHLEN];
+
+		ret = unpack(nodospath(path, filelist[filepos].name),
+			NULL, arg, 0, 1);
+#endif
+	}
 	if (ret < 0) return(warning_bell(arg));
 	if (!ret) return(1);
 	return(4);
@@ -1383,7 +1403,16 @@ char *arg;
 
 	if (isdir(&(filelist[filepos]))) return(warning_bell(arg));
 	if (archivefile) ret = unpack(archivefile, NULL, NULL, 1, 0);
-	else ret = unpack(filelist[filepos].name, NULL, NULL, 1, 1);
+	else {
+#if	MSDOS || defined (_NODOSDRIVE)
+		ret = unpack(filelist[filepos].name, NULL, NULL, 1, 1);
+#else
+		char path[MAXPATHLEN];
+
+		ret = unpack(nodospath(path, filelist[filepos].name),
+			NULL, NULL, 1, 1);
+#endif
+	}
 	if (ret <= 0) {
 		if (ret < 0) warning_bell(arg);
 		return(3);
@@ -1431,7 +1460,7 @@ char *arg;
 		locate(0, LINFO);
 		putterm(l_clear);
 		kanjiputs(ATTRM_K);
-		if (selectstr(&flag, 2, 35, str, val) == ESC) return(1);
+		if (selectstr(&flag, 2, 35, str, val) == K_ESC) return(1);
 	}
 	else {
 #if	!MSDOS
@@ -1447,9 +1476,7 @@ char *arg;
 	while ((n = inputattr(&(filelist[i]), flag)) < 0) warning(0, ILTMS_K);
 	if (!n) return(2);
 
-	if (mark > 0) applyfile(setattr, NULL);
-	else if (setattr(filelist[filepos].name) < 0)
-		warning(-1, filelist[filepos].name);
+	applyfile(setattr, NULL);
 	return(4);
 }
 
