@@ -32,6 +32,7 @@
 #define	C_EOF		('Z' & 037)
 #else	/* !MSDOS */
 #include <sys/time.h>
+#include <sys/file.h>
 #include <sys/param.h>
 # ifdef	USEDIRECT
 # include <sys/dir.h>
@@ -67,6 +68,21 @@
 #define	S_IWUSR	00200
 #endif
 
+#ifndef	L_SET
+# ifdef	SEEK_SET
+# define	L_SET	SEEK_SET
+# else
+# define	L_SET	0
+# endif
+#endif
+#ifndef	L_INCR
+# ifdef	SEEK_CUR
+# define	L_INCR	SEEK_CUR
+# else
+# define	L_INCR	1
+# endif
+#endif
+
 #include "term.h"
 #include "pathname.h"
 #include "system.h"
@@ -211,19 +227,6 @@ struct filestat_t {
 	time_t mtim;
 	time_t atim;
 };
-static char dirsort[sizeof("NSEDGA")] = "";
-static char dirattr[sizeof("DRHSA")] = "";
-static int dirtype = '\0';
-static int dirflag = 0;
-#define	DF_LOWER	001
-#define	DF_PAUSE	002
-#define	DF_SUBDIR	004
-static int copyflag = 0;
-#define	CF_BINARY	001
-#define	CF_TEXT		002
-#define	CF_VERIFY	004
-#define	CF_NOCONFIRM	010
-#define	CF_VERBOSE	020
 
 #ifndef	O_BINARY
 #define	O_BINARY	0
@@ -275,6 +278,20 @@ static char *doserrstr[] = {
 };
 #define	DOSERRSIZ	((int)(sizeof(doserrstr) / sizeof(char *)))
 
+static char dirsort[sizeof("NSEDGA")] = "";
+static char dirattr[sizeof("DRHSA")] = "";
+static int dirtype = '\0';
+static int dirflag = 0;
+#define	DF_LOWER	001
+#define	DF_PAUSE	002
+#define	DF_SUBDIR	004
+static int copyflag = 0;
+#define	CF_BINARY	001
+#define	CF_TEXT		002
+#define	CF_VERIFY	004
+#define	CF_NOCONFIRM	010
+#define	CF_VERBOSE	020
+
 
 #ifndef	FD
 static char *NEAR ascnumeric(buf, n, digit, max)
@@ -483,7 +500,7 @@ int (NEAR *getoptcmd)__P_((int, char *[]));
 	argv[0] = strdup2(cmd);
 	argv[1] = strdup2(cp);
 	argv[2] = NULL;
-	argc = evalifs(argc, &argv, IFS_SET, 1);
+	argc = evalifs(argc, &argv, IFS_SET);
 
 	er = 0;
 	if ((n = (*getoptcmd)(argc, argv)) < 0) er++;
@@ -548,6 +565,7 @@ char *argv[];
 							r = 1;
 							break;
 						}
+/*FALLTHRU*/
 					default:
 						doserror(cp, ER_REQPARAM);
 						return(-1);
@@ -591,6 +609,7 @@ char *argv[];
 							r = 1;
 							break;
 						}
+/*FALLTHRU*/
 					default:
 						doserror(cp, ER_REQPARAM);
 						return(-1);
@@ -1322,7 +1341,7 @@ char *file, *src;
 		key = -1;
 		for (;;) {
 			c = inputkey();
-			if (c == CR && key >= 0) break;
+			if (c == K_CR && key >= 0) break;
 			if (!strchr("ynaYNA", c)) continue;
 			key = toupper2(c);
 			fputc(c, stderr);
@@ -1405,7 +1424,7 @@ int sbin, dbin, dfd;
 		return(-1);
 	if (dfd >= 0) {
 		fd2 = dfd;
-		ofs = lseek(fd2, 0L, SEEK_CUR);
+		ofs = lseek(fd2, 0L, L_INCR);
 	}
 	else if ((fd2 = writeopen(dest, src)) <= 0) {
 		Xclose(fd1);
@@ -1433,8 +1452,8 @@ int sbin, dbin, dfd;
 		}
 
 		if (i < 0 || !(copyflag & CF_VERIFY)
-		|| lseek(fd1, 0L, SEEK_SET) < 0
-		|| lseek(fd2, ofs, SEEK_SET) < 0) break;
+		|| lseek(fd1, 0L, L_SET) < 0
+		|| lseek(fd2, ofs, L_SET) < 0) break;
 		for (;;) {
 			while ((i = textread(fd1, buf, BUFSIZ, sbin)) < 0
 			&& errno == EINTR);
