@@ -86,7 +86,7 @@ extern char *_mtrace_file;
 #define	ADJTTY		0
 #define	USEGETCURSOR	0
 #define	WINDOWS		1
-#define	COLUMNS		2
+#define	DEFCOLUMNS	2
 #define	MINFILENAME	12
 #define	HISTSIZE	50
 #define	DIRHIST		50
@@ -157,7 +157,7 @@ extern char *_mtrace_file;
 /****************************************************************
  *	Screen layout parameter					*
  ****************************************************************/
-#define	FILEPERLINE	(columns)
+#define	FILEPERLINE	(curcolumns)
 #ifndef	_NOSPLITWIN
 #define	FILEPERLOW	((n_line - WHEADER - WFOOTER + 1) / windows - 1)
 #else
