@@ -44,6 +44,7 @@ extern char *tmpfilename;
 
 #define	DOSDIRENT		32
 #define	LFNENTSIZ		13
+#define	MAXTMPNAMLEN		8
 #ifndef	O_BINARY
 #define	O_BINARY		0
 #endif
@@ -61,6 +62,7 @@ static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
 #ifndef	_NOWRITEFS
 static int NEAR k_strlen __P_((char *));
+static int NEAR saferename __P_((char *, char *));
 static char *NEAR maketmpfile __P_((int, int, char *, char *));
 #if	!MSDOS
 static char *NEAR getentnum __P_((char *, int, int));
@@ -78,6 +80,18 @@ static int dosdrv = -1;
 #endif
 
 
+#if	!MSDOS && !defined (_NODOSDRIVE)
+char *nodospath(path, file)
+char *path, *file;
+{
+	if (!_dospath(file)) return(file);
+	path[0] = '.';
+	path[1] = _SC_;
+	strcpy(path + 2, file);
+	return(path);
+}
+#endif
+
 #if	MSDOS
 int logical_access(mode)
 u_short mode;
@@ -166,15 +180,21 @@ namelist *list;
 	if (stat2(list -> name, &st) < 0) return(-1);
 	memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 #else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
-	if (dospath(list -> name, NULL)) {
-		if (stat2(list -> name, &st) < 0) return(-1);
+	char *cp;
+
+# ifdef	_NODOSDRIVE
+	cp = list -> name;
+# else
+	char path[MAXPATHLEN];
+
+	cp = nodospath(path, list -> name);
+	if (dospath(cp, NULL)) {
+		if (stat2(cp, &st) < 0) return(-1);
 		memcpy((char *)&lst, (char *)&st, sizeof(struct stat));
 	}
 	else
 # endif
-	if (Xlstat(list -> name, &lst) < 0
-	|| stat2(list -> name, &st) < 0) return(-1);
+	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
 #endif	/* !MSDOS */
 	list -> flags = 0;
 	if ((st.st_mode & S_IFMT) == S_IFDIR) list -> flags |= F_ISDIR;
@@ -252,8 +272,10 @@ CONST VOID_P vp2;
 
 	switch (sorton & 7) {
 		case 5:
-			tmp = strlen(list1 -> name) - strlen(list2 -> name);
+			tmp = (int)strlen(list1 -> name)
+				- (int)strlen(list2 -> name);
 			if (tmp != 0) break;
+/*FALLTHRU*/
 		case 1:
 			tmp = strpathcmp2(list1 -> name, list2 -> name);
 			break;
@@ -707,9 +729,9 @@ char *dir;
 	*path = (no >= 'A' && no <= 'Z') ? toupper2(*path) : tolower2(*path);
 #endif
 	strcpy(strcatdelim(path), tmpfilename);
-	if (_Xmkdir(path, 0777) < 0 && errno != EEXIST) return(-1);
+	if (_Xmkdir(path, 0755) < 0 && errno != EEXIST) return(-1);
 	strcpy((cp = strcatdelim(path)), dir);
-	if (_Xmkdir(path, 0777) < 0 && errno != EEXIST) {
+	if (_Xmkdir(path, 0755) < 0 && errno != EEXIST) {
 		*(--cp) = '\0';
 		no = errno;
 		if (_Xrmdir(path) < 0
@@ -739,6 +761,71 @@ char *dir;
 	return(0);
 }
 
+int mktmpfile(buf, dir)
+char *buf, *dir;
+{
+	char *cp, path[MAXPATHLEN];
+	int i, fd;
+
+	strcpy(path, dir);
+	if (mktmpdir(path) < 0) return(-1);
+	cp = strcatdelim(path);
+	memset(cp, '\0', sizeof(path) - (cp - path));
+	cp[0] = '_';
+
+	for (;;) {
+		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL, 0644);
+		if (fd >= 0) {
+			strcpy(buf, path);
+			return(fd);
+		}
+		if (errno != EEXIST) return(-1);
+
+		for (i = 0; i < MAXTMPNAMLEN; i++) {
+			if (!cp[i]) cp[i] = '_';
+			else if (cp[i] == '9') {
+				cp[i] = '_';
+				continue;
+			}
+			else if (cp[i] == '_') cp[i] = 'a';
+#if	MSDOS
+			else if (cp[i] == 'z') cp[i] = '0';
+#else
+			else if (cp[i] == 'z') cp[i] = 'A';
+			else if (cp[i] == 'Z') cp[i] = '0';
+#endif
+			else cp[i]++;
+			break;
+		}
+		if (i >= MAXTMPNAMLEN) break;
+	}
+	if (cp > path) *(--cp) = '\0';
+	rmtmpdir(path);
+	return(-1);
+}
+
+int rmtmpfile(path)
+char *path;
+{
+	char *cp;
+	int ret;
+
+	ret = 0;
+	if (Xunlink(path) != 0 && errno != ENOENT) ret = -1;
+	else if (!(cp = strrdelim(path, 0)));
+#if	MSDOS
+	else if (cp == path + 2 && isalpha(path[0]) && path[1] == ':');
+#endif
+	else if (cp != path) {
+		*cp = '\0';
+		if (rmtmpdir(path) < 0
+		&& errno != ENOTEMPTY && errno != EEXIST && errno != EACCES)
+			ret = -1;
+	}
+	free(path);
+	return(ret);
+}
+
 VOID removetmp(dir, subdir, file)
 char *dir, *subdir, *file;
 {
@@ -759,6 +846,11 @@ char *dir, *subdir, *file;
 	}
 	if (file) {
 		if (*file) {
+#if	!MSDOS && !defined (_NODOSDRIVE)
+			char path[MAXPATHLEN];
+
+			file = nodospath(path, file);
+#endif
 			if (Xunlink(file) < 0) {
 				warning(-1, file);
 				subdir = NULL;
@@ -833,6 +925,9 @@ int single;
 {
 	struct stat st;
 	char *cp, path[MAXPATHLEN];
+# if	!MSDOS
+	char tmp[MAXPATHLEN];
+# endif
 	int i, drive;
 
 	if (!(drive = dospath2(dir))) return(0);
@@ -852,7 +947,13 @@ int single;
 		strcpy(cp, filelist[filepos].name);
 		st.st_mode = filelist[filepos].st_mode;
 		st.st_atime = st.st_mtime = filelist[filepos].st_mtim;
-		if (cpfile(filelist[filepos].name, path, &st, NULL) < 0) {
+# if	MSDOS
+		if (cpfile(filelist[filepos].name, path, &st, NULL) < 0)
+# else
+		if (cpfile(nodospath(tmp, filelist[filepos].name),
+		path, &st, NULL) < 0)
+# endif
+		{
 			*(--cp) = '\0';
 			removetmp(path, NULL, NULL);
 			return(-1);
@@ -863,7 +964,13 @@ int single;
 		strcpy(cp, filelist[i].name);
 		st.st_mode = filelist[i].st_mode;
 		st.st_atime = st.st_mtime = filelist[i].st_mtim;
-		if (cpfile(filelist[i].name, path, &st, NULL) < 0) {
+# if	MSDOS
+		if (cpfile(filelist[i].name, path, &st, NULL) < 0)
+# else
+		if (cpfile(nodospath(tmp, filelist[i].name),
+		path, &st, NULL) < 0)
+# endif
+		{
 			*(--cp) = '\0';
 			removetmp(path, NULL, "");
 			return(-1);
@@ -884,10 +991,16 @@ char *file;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
+# if	!MSDOS
+	char tmp[MAXPATHLEN];
+# endif
 
 	path[0] = drive;
 	path[1] = ':';
 	strcpy(path + 2, file);
+# if	!MSDOS
+	file = nodospath(tmp, file);
+# endif
 	waitmes();
 	if (Xlstat(file, &st) < 0 || cpfile(file, path, &st, NULL) < 0)
 		return(-1);
@@ -919,6 +1032,21 @@ char *s;
 	return(i);
 }
 
+static int NEAR saferename(from, to)
+char *from, *to;
+{
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	char fpath[MAXPATHLEN], tpath[MAXPATHLEN];
+#endif
+
+	if (!strpathcmp(from, to)) return(0);
+#if	!MSDOS && !defined (_NODOSDRIVE)
+	from = nodospath(fpath, from);
+	to = nodospath(tpath, to);
+#endif
+	return(Xrename(from, to));
+}
+
 static char *NEAR maketmpfile(len, dos, tmpdir, old)
 int len, dos;
 char *tmpdir, *old;
@@ -943,14 +1071,12 @@ char *tmpdir, *old;
 					if (isexist(fname));
 					else
 #endif
-					{
-						if (rename2(old, fname) >= 0)
-							return(fname);
+					if (saferename(old, fname) >= 0)
+						return(fname);
 #if	MSDOS
-						if (errno != EACCES)
+					else if (errno == EACCES);
 #endif
-						break;
-					}
+					else break;
 				}
 				else {
 					if ((fd = Xopen(fname,
@@ -1030,7 +1156,7 @@ int fnamp;
 		if (isdotdir(dp -> d_name)) continue;
 		else {
 			strcpy(path + fnamp, dp -> d_name);
-			if (rename2(path, dp -> d_name) < 0) error(path);
+			if (saferename(path, dp -> d_name) < 0) error(path);
 		}
 	}
 	Xclosedir(dirp);
@@ -1119,7 +1245,7 @@ int fs;
 		}
 		else {
 			strcpy(path + fnamp, dp -> d_name);
-			if (rename2(dp -> d_name, path) < 0) {
+			if (saferename(dp -> d_name, path) < 0) {
 				Xclosedir(dirp);
 				warning(-1, dp -> d_name);
 				restorefile(tmpdir, path, fnamp);
@@ -1177,7 +1303,7 @@ int fs;
 		}
 #endif	/* !MSDOS */
 		strcpy(path + fnamp, filelist[i].name);
-		if (rename2(path, filelist[i].name) < 0) error(path);
+		if (saferename(path, filelist[i].name) < 0) error(path);
 #if	!MSDOS
 		totalent += realdirsiz(filelist[i].name);
 		ptr++;
@@ -1191,7 +1317,7 @@ int fs;
 	free(tmpdir);
 	tmpdir = cp;
 	strcatdelim2(path, tmpdir, filelist[top].name);
-	if (rename2(path, filelist[top].name) < 0) error(path);
+	if (saferename(path, filelist[top].name) < 0) error(path);
 	restorefile(tmpdir, path, fnamp);
 
 #if	!MSDOS
