@@ -7,6 +7,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
+#include <fcntl.h>
 #include "machine.h"
 
 #ifndef NOUNISTDH
@@ -17,8 +18,13 @@
 #include <stdlib.h>
 #endif
 
+#ifdef	USETIMEH
+#include <time.h>
+#endif
+
 #include "term.h"
 #include "kctype.h"
+#include "pathname.h"
 
 #ifdef	USESTDARGH
 #include <stdarg.h>
@@ -26,12 +32,38 @@
 #include <varargs.h>
 #endif
 
+#if	MSDOS
+#include <io.h>
+#include "unixemu.h"
+#else
+#include <sys/file.h>
+#include <sys/param.h>
+#endif
+
+#ifndef	L_SET
+# ifdef	SEEK_SET
+# define	L_SET	SEEK_SET
+# else
+# define	L_SET	0
+# endif
+#endif
+
 #define	ASCII		000
 #define	KANA		001
 #define	KANJI		002
 #define	JKANA		004
 #define	MAXPRINTBUF	1023
 #define	SJ_UDEF		0x81ac	/* GETA */
+#define	UNICODETBL	"fd-unicd.tbl"
+#define	MINUNICODE	0x00a7
+#define	MAXUNICODE	0xffe5
+#define	MINKANJI	0x8140
+#define	MAXKANJI	0xfc4b
+#define	jcnv(c, io)	((((io) == L_FNAME) && ((c) == '/')) ? ' ' : (c))
+#define	jdecnv(c, io)	((((io) == L_FNAME) && ((c) == ' ')) ? '/' : (c))
+#ifndef	O_BINARY
+#define	O_BINARY	0
+#endif
 
 typedef struct _kconv_t {
 	u_short start;
@@ -40,33 +72,34 @@ typedef struct _kconv_t {
 } kconv_t;
 
 extern char *malloc2 __P_((ALLOC_T));
+extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *strncpy3 __P_((char *, char *, int *, int));
 extern char *strstr2 __P_((char *, char *));
 
 #if	!MSDOS && (!defined (_NOKANJICONV) \
 || (!defined (_NODOSDRIVE) && defined (CODEEUC)))
-static VOID NEAR sjis2jis __P_((char *, u_char *));
-static VOID NEAR jis2sjis __P_((char *, u_char *));
+static VOID NEAR sj2j __P_((char *, u_char *));
+static VOID NEAR j2sj __P_((char *, u_char *));
 #endif
 #if	!MSDOS && !defined (_NOKANJICONV)
 static int NEAR toenglish __P_((char *, u_char *, int));
-static int NEAR tojis7 __P_((char *, u_char *, int, int, int));
-static int NEAR fromjis7 __P_((char *, u_char *, int, int, int));
-#endif	/* !MSDOS && !_NOKANJICONV */
-#if	!MSDOS && !defined (_NOKANJIFCONV)
-static int NEAR tojis8 __P_((char *, u_char *, int, int, int));
-static int NEAR fromjis8 __P_((char *, u_char *, int, int, int));
-static int NEAR tojunet __P_((char *, u_char *, int, int, int));
-static int NEAR fromjunet __P_((char *, u_char *, int, int, int));
+static int NEAR tojis7 __P_((char *, u_char *, int, int, int, int));
+static int NEAR fromjis __P_((char *, u_char *, int, int));
+# if	defined (FD) && (FD >= 2)
+static int NEAR tojis8 __P_((char *, u_char *, int, int, int, int));
+static int NEAR tojunet __P_((char *, u_char *, int, int, int, int));
+static int NEAR toutf8 __P_((char *, u_char *, int));
+static int NEAR fromutf8 __P_((char *, u_char *, int));
+#  ifndef	_NOKANJIFCONV
 static int NEAR bin2hex __P_((char *, int));
 static int NEAR tohex __P_((char *, u_char *, int));
 static int NEAR fromhex __P_((char *, u_char *, int));
 static int NEAR bin2cap __P_((char *, int));
 static int NEAR tocap __P_((char *, u_char *, int));
 static int NEAR fromcap __P_((char *, u_char *, int));
-#endif	/* !MSDOS && !_NOKANJIFCONV */
-#if	!MSDOS && !defined (_NOKANJICONV)
-static char *NEAR _kanjiconv __P_((char *, char *, int, int, int, int *));
+#  endif	/* !_NOKANJIFCONV */
+# endif	/* FD && (FD >= 2) */
+static char *NEAR _kanjiconv __P_((char *, char *, int, int, int, int *, int));
 #endif	/* !MSDOS && !_NOKANJICONV */
 
 #if	!MSDOS && !defined (_NOKANJICONV)
@@ -76,7 +109,7 @@ int inputkcode = 0;
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
 int outputkcode = 0;
 #endif
-#if	!MSDOS && !defined (_NOKANJIKCONV)
+#if	!MSDOS && !defined (_NOKANJIFCONV)
 int fnamekcode = 0;
 #endif
 #ifndef	LSI_C
@@ -115,7 +148,43 @@ u_char kctypetable[256] = {
 	0013, 0013, 0013, 0013, 0013, 0010, 0010,    0
 };
 #endif	/* !LSI_C */
-
+#if	(!MSDOS && !defined (_NOKANJICONV)) || !defined (_NODOSDRIVE)
+char *unitblpath = NULL;
+static u_char *unitblbuf = NULL;
+static u_short unitblent = 0;
+static kconv_t rsjistable[] = {
+	{0x8470, 0x8440, 0x0f},		/* strange Russian char */
+	{0x8480, 0x844f, 0x12},		/* Why they converted ? */
+#define	EXCEPTRUSS	2
+	{0x8754, 0xfa4a, 0x0a},
+	{0x8782, 0xfa59, 0x01},
+	{0x8784, 0xfa5a, 0x01},
+	{0x878a, 0xfa58, 0x01},
+	{0x8790, 0x81e0, 0x01},
+	{0x8791, 0x81df, 0x01},
+	{0x8792, 0x81e7, 0x01},
+	{0x8795, 0x81e3, 0x01},
+	{0x8796, 0x81db, 0x01},
+	{0x8797, 0x81da, 0x01},
+	{0x879a, 0x81e6, 0x01},
+	{0x879b, 0x81bf, 0x01},
+	{0x879c, 0x81be, 0x01},
+	{0xed40, 0xfa5c, 0x23},
+	{0xed63, 0xfa80, 0x1c},
+	{0xed80, 0xfa9c, 0x61},
+	{0xede1, 0xfb40, 0x1c},
+	{0xee40, 0xfb5c, 0x23},
+	{0xee63, 0xfb80, 0x1c},
+	{0xee80, 0xfb9c, 0x61},
+	{0xeee1, 0xfc40, 0x0c},
+	{0xeeef, 0xfa40, 0x0a},
+	{0xeef9, 0x81ca, 0x01},
+	{0xeefa, 0xfa55, 0x03},
+	{0xfa54, 0x81ca, 0x01},
+	{0xfa5b, 0x81e6, 0x01},
+};
+#define	RSJISTBLSIZ	((int)(sizeof(rsjistable) / sizeof(kconv_t)))
+#endif	/* (!MSDOS && !_NOKANJICONV) || !_NODOSDRIVE */
 #if	!MSDOS && (!defined (_NOKANJICONV) \
 || (!defined (_NODOSDRIVE) && defined (CODEEUC)))
 static kconv_t convtable[] = {
@@ -136,7 +205,7 @@ static kconv_t convtable[] = {
 	{0xfb9c, 0xee80, 0x61},
 	{0xfc40, 0xeee1, 0x0c}
 };
-#define	CNVTBLSIZ	(sizeof(convtable) / sizeof(kconv_t))
+#define	CNVTBLSIZ	((int)(sizeof(convtable) / sizeof(kconv_t)))
 static u_char sj2jtable1[256] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
@@ -222,7 +291,7 @@ static u_char j2sjtable2[128] = {
 	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,    0
 };
 #endif	/* !MSDOS && (!_NOKANJICONV || (!_NODOSDRIVE && CODEEUC)) */
-#if	!MSDOS && !defined (_NOKANJIFCONV)
+#if	!MSDOS && !defined (_NOKANJIFCONV) && defined (FD) && (FD >= 2)
 static u_char hctypetable[256] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
@@ -280,7 +349,7 @@ static u_char b2htable[256] = "0123456789abcdef";
 #define	iscap(s, i)	(((s)[i] == CAPTAG) \
 			&& (hctypetable[(u_char)(s)[(i) + 1]] & HC_CAP) \
 			&& (hctypetable[(u_char)(s)[(i) + 2]] & HC_HEX))
-#endif	/* !MSDOS && !_NOKANJIFCONV */
+#endif	/* !MSDOS && !_NOKANJIFCONV && FD && FD >= 2 */
 
 int onkanji1 __P_((char *, int));
 #if	(!MSDOS && !defined (_NOKANJICONV)) \
@@ -295,12 +364,20 @@ char *mesconv __P_((char *, char *));
 int sjis2ujis __P_((char *, u_char *, int));
 int ujis2sjis __P_((char *, u_char *, int));
 #endif
+#if	(!MSDOS && defined (FD) && (FD >= 2) && !defined (_NOKANJICONV)) \
+|| !defined (_NODOSDRIVE)
+VOID readunitable __P_((VOID_A));
+VOID discardunitable __P_((VOID_A));
+u_short unifysjis __P_((u_short, int));
+u_short cnvunicode __P_((u_short, int));
+#endif
 #if	!MSDOS && !defined (_NOKANJICONV)
-int kanjiconv __P_((char *, char *, int, int, int));
-#endif
-#if	!MSDOS && !defined (_NOKANJIFCONV)
+int kanjiconv __P_((char *, char *, int, int, int, int));
+# ifndef	_NOKANJIFCONV
 char *kanjiconv2 __P_((char *, char *, int, int, int));
+# endif
 #endif
+int kanjifputs __P_((char *, FILE *));
 int kanjiputs __P_((char *));
 int kanjiprintf __P_((CONST char *, ...));
 int kanjiputs2 __P_((char *, int, int));
@@ -340,24 +417,25 @@ int io;
 	int ret;
 
 	if (!s) ret = NOCNV;
-# if	!MSDOS && !defined (_NOKANJIFCONV)
+# if	!MSDOS && !defined (_NOKANJICONV)
+#  ifndef	_NOKANJIFCONV
 	else if (io == L_FNAME && (strstr2(s, "HEX") || strstr2(s, "hex")))
 		ret = HEX;
 	else if (io == L_FNAME && (strstr2(s, "CAP") || strstr2(s, "cap")))
 		ret = CAP;
-# endif	/* !MSDOS && !_NOKANJIFCONV */
-# if	!MSDOS && !defined (_NOKANJICONV)
+#  endif	/* !_NOKANJIFCONV */
 	else if (strstr2(s, "SJIS") || strstr2(s, "sjis")) ret = SJIS;
 	else if (strstr2(s, "EUC") || strstr2(s, "euc")
 	|| strstr2(s, "UJIS") || strstr2(s, "ujis")) ret = EUC;
-# ifndef	_NOKANJIFCONV
+#  if	defined (FD) && (FD >= 2)
+	else if (strstr2(s, "UTF8") || strstr2(s, "utf8")) ret = UTF8;
 	else if (strstr2(s, "OJUNET") || strstr2(s, "ojunet"))
 		ret = O_JUNET;
 	else if (strstr2(s, "OJIS8") || strstr2(s, "ojis8")) ret = O_JIS8;
 	else if (strstr2(s, "OJIS") || strstr2(s, "ojis")) ret = O_JIS7;
 	else if (strstr2(s, "JUNET") || strstr2(s, "junet")) ret = JUNET;
 	else if (strstr2(s, "JIS8") || strstr2(s, "jis8")) ret = JIS8;
-# endif	/* !_NOKANJIFCONV */
+#  endif	/* FD && (FD >= 2) */
 	else if (strstr2(s, "JIS") || strstr2(s, "jis")) ret = JIS7;
 # endif	/* !MSDOS && !_NOKANJICONV */
 # ifndef	_NOENGMES
@@ -390,7 +468,7 @@ char *jpn, *eng;
 
 #if	!MSDOS && (!defined (_NOKANJICONV) \
 || (!defined (_NODOSDRIVE) && defined (CODEEUC)))
-static VOID NEAR sjis2jis(buf, s)
+static VOID NEAR sj2j(buf, s)
 char *buf;
 u_char *s;
 {
@@ -422,7 +500,7 @@ u_char *s;
 	buf[1] = j2;
 }
 
-static VOID NEAR jis2sjis(buf, s)
+static VOID NEAR j2sj(buf, s)
 char *buf;
 u_char *s;
 {
@@ -463,7 +541,7 @@ int max;
 			buf[j] = s[i];
 		}
 		else if (issjis1(s[i]) && issjis2(s[i + 1])) {
-			sjis2jis(&(buf[j]), &(s[i++]));
+			sj2j(&(buf[j]), &(s[i++]));
 			buf[j++] |= 0x80;
 			buf[j] |= 0x80;
 		}
@@ -481,13 +559,207 @@ int max;
 
 	for (i = j = 0; s[i] && j < max - 1; i++, j++) {
 		if (isekana(s, i)) buf[j] = s[++i];
-		else if (iseuc(s[i])) jis2sjis(&(buf[j++]), &(s[i++]));
+		else if (iseuc(s[i]) && iseuc(s[i + 1]))
+			j2sj(&(buf[j++]), &(s[i++]));
 		else buf[j] = s[i];
 	}
 	return(j);
 }
 #endif	/* !MSDOS && (!_NOKANJICONV || (!_NODOSDRIVE && CODEEUC)) */
 
+#if	(!MSDOS && defined (FD) && (FD >= 2) && !defined (_NOKANJICONV)) \
+|| !defined (_NODOSDRIVE)
+VOID readunitable(VOID_A)
+{
+	char path[MAXPATHLEN];
+	u_char buf[2];
+	long size;
+	int fd;
+
+	if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
+	else strcatdelim2(path, unitblpath, UNICODETBL);
+
+	if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0) return;
+	if (read(fd, buf, 2) != 2) {
+		close(fd);
+		return;
+	}
+	unitblent = (((u_short)(buf[1]) << 8) | buf[0]);
+	size = (long)unitblent * 4;
+
+	unitblbuf = (u_char *)realloc2(unitblbuf, size);
+	if (read(fd, unitblbuf, size) != size) discardunitable();
+	close(fd);
+}
+
+VOID discardunitable(VOID_A)
+{
+	if (unitblbuf) {
+		free(unitblbuf);
+		unitblbuf = NULL;
+	}
+}
+
+u_short unifysjis(wc, russ)
+u_short wc;
+int russ;
+{
+	int i;
+
+	for (i = ((russ) ? 0 : EXCEPTRUSS); i < RSJISTBLSIZ; i++)
+		if (wc >= rsjistable[i].start
+		&& wc < rsjistable[i].start + rsjistable[i].range)
+			break;
+	if (i < RSJISTBLSIZ) {
+		wc -= rsjistable[i].start;
+		wc += rsjistable[i].cnv;
+	}
+	return(wc);
+}
+
+u_short cnvunicode(wc, encode)
+u_short wc;
+int encode;
+{
+	static int fd = -1;
+	u_char buf[4];
+	char path[MAXPATHLEN];
+	u_short r, w, min, max, ofs;
+
+	if (encode < 0) {
+		if (fd >= 0) close(fd);
+		fd = -1;
+		return(0);
+	}
+
+	if (encode) {
+		r = 0xff00;
+		if (wc < 0x80) return(wc);
+		if (wc > 0xa0 && wc <= 0xdf) return(0xff00 | (wc - 0x40));
+		if (wc >= 0x8260 && wc <= 0x8279)
+			return(0xff00 | (wc - 0x8260 + 0x21));
+		if (wc >= 0x8281 && wc <= 0x829a)
+			return(0xff00 | (wc - 0x8281 + 0x41));
+		if (wc < MINKANJI || wc > MAXKANJI) return(r);
+	}
+	else {
+		r = '?';
+		switch (wc & 0xff00) {
+			case 0:
+				if ((wc & 0xff) < 0x80) return(wc);
+				break;
+			case 0xff00:
+				w = wc & 0xff;
+				if (w > 0x20 && w <= 0x3a)
+					return(w + 0x8260 - 0x21);
+				if (w > 0x40 && w <= 0x5a)
+					return(w + 0x8281 - 0x41);
+				if (w > 0x60 && w <= 0x9f)
+					return(w + 0x40);
+				break;
+			default:
+				break;
+		}
+		if (wc < MINUNICODE || wc > MAXUNICODE) return(r);
+	}
+
+	if (unitblbuf) {
+		u_char *cp;
+
+		if (encode) {
+			wc = unifysjis(wc, 0);
+			cp = unitblbuf;
+			for (ofs = 0; ofs < unitblent; ofs++) {
+				w = (((u_short)(cp[3]) << 8) | cp[2]);
+				if (wc == w) {
+					r = (((u_short)(cp[1]) << 8) | cp[0]);
+					break;
+				}
+				cp += 4;
+			}
+		}
+		else {
+			min = 0;
+			max = unitblent + 1;
+			ofs = unitblent / 2 + 1;
+			for (;;) {
+				if (ofs == min || ofs == max) break;
+				cp = &(unitblbuf[(ofs - 1) * 4]);
+				w = (((u_short)(cp[1]) << 8) | cp[0]);
+				if (wc == w) {
+					r = (((u_short)(cp[3]) << 8) | cp[2]);
+					break;
+				}
+				else if (wc < w) {
+					max = ofs;
+					ofs = (ofs + min) / 2;
+				}
+				else {
+					min = ofs;
+					ofs = (ofs + max) / 2;
+				}
+			}
+		}
+
+		return(r);
+	}
+
+	if (fd < 0) {
+		if (!unitblpath || !*unitblpath) strcpy(path, UNICODETBL);
+		else strcatdelim2(path, unitblpath, UNICODETBL);
+
+		if ((fd = open(path, O_RDONLY | O_BINARY, 0600)) < 0)
+			return(r);
+		if (!unitblent) {
+			if (read(fd, buf, 2) != 2) {
+				close(fd);
+				fd = -1;
+				return(r);
+			}
+			unitblent = (((u_short)(buf[1]) << 8) | buf[0]);
+		}
+	}
+
+	if (encode) {
+		if (lseek(fd, (off_t)2, L_SET) < 0) return(r);
+		wc = unifysjis(wc, 0);
+		for (ofs = 0; ofs < unitblent; ofs++) {
+			if (read(fd, buf, 4) != 4) break;
+			w = (((u_short)(buf[3]) << 8) | buf[2]);
+			if (wc == w) {
+				r = (((u_short)(buf[1]) << 8) | buf[0]);
+				break;
+			}
+		}
+	}
+	else {
+		min = 0;
+		max = unitblent + 1;
+		ofs = unitblent / 2 + 1;
+		for (;;) {
+			if (ofs == min || ofs == max) break;
+			if (lseek(fd, (off_t)(ofs - 1) * 4 + 2, L_SET) < 0
+			|| read(fd, buf, 4) != 4) break;
+			w = (((u_short)(buf[1]) << 8) | buf[0]);
+			if (wc == w) {
+				r = (((u_short)(buf[3]) << 8) | buf[2]);
+				break;
+			}
+			else if (wc < w) {
+				max = ofs;
+				ofs = (ofs + min) / 2;
+			}
+			else {
+				min = ofs;
+				ofs = (ofs + max) / 2;
+			}
+		}
+	}
+
+	return(r);
+}
+#endif	/* (!MSDOS && FD && (FD >= 2) && !_NOKANJICONV) || !_NODOSDRIVE */
+
 #if	!MSDOS && !defined (_NOKANJICONV)
 static int NEAR toenglish(buf, s, max)
 char *buf;
@@ -515,10 +787,10 @@ int max;
 	return(j);
 }
 
-static int NEAR tojis7(buf, s, max, knj, asc)
+static int NEAR tojis7(buf, s, max, knj, asc, io)
 char *buf;
 u_char *s;
-int max, knj, asc;
+int max, knj, asc, io;
 {
 	int i, j, mode;
 
@@ -533,6 +805,7 @@ int max, knj, asc;
 			if (!(mode & KANA)) buf[j++] = '\016';
 			mode |= KANA;
 			buf[j] = s[i] & ~0x80;
+			buf[j] = jcnv(buf[j], io);
 			continue;
 		}
 		if (mode & KANA) buf[j++] = '\017';
@@ -548,8 +821,9 @@ int max, knj, asc;
 			buf[j++] = s[i++] & ~0x80;
 			buf[j] = s[i] & ~0x80;
 # else
-			sjis2jis(&(buf[j++]), &(s[i++]));
+			sj2j(&(buf[j++]), &(s[i++]));
 # endif
+			buf[j] = jcnv(buf[j], io);
 		}
 		else {
 			if (mode & KANJI) {
@@ -570,10 +844,10 @@ int max, knj, asc;
 	return(j);
 }
 
-static int NEAR fromjis7(buf, s, max, knj, asc)
+static int NEAR fromjis(buf, s, max, io)
 char *buf;
 u_char *s;
-int max, knj, asc;
+int max, io;
 {
 	int i, j, mode;
 
@@ -587,24 +861,54 @@ int max, knj, asc;
 			mode &= ~KANA;
 			break;
 		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode |= KANJI;
-				i += 2;
-				break;
+			if (s[i + 1] == '$') {
+				if (s[i + 2] == '@' || s[i + 2] == 'B') {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 2;
+					break;
+				}
+				else if (s[i + 2] == '('
+				&& (s[i + 3] == '@' || s[i + 3] == 'B'
+				|| s[i + 3] == 'O')) {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 3;
+					break;
+				}
 			}
-			else if (s[i + 1] == '(' && s[i + 2] == asc) {
-				mode &= ~KANJI;
-				i += 2;
-				break;
+			else if (s[i + 1] == '(') {
+				if (s[i + 2] == 'J' || s[i + 2] == 'B'
+				|| s[i + 2] == 'H') {
+					mode &= ~(KANJI | JKANA);
+					i += 2;
+					break;
+				}
+				else if (s[i + 2] == 'I') {
+					mode &= ~KANJI;
+					mode |= JKANA;
+					i += 2;
+					break;
+				}
 			}
+			else if (s[i + 1] == '&') {
+				if (s[i + 2] == '@' && s[i + 3] == '\033'
+				&& s[i + 4] == '$' && s[i + 5] == 'B') {
+					mode &= ~JKANA;
+					mode |= KANJI;
+					i += 5;
+					break;
+				}
+			}
+/*FALLTHRU*/
 		default:
-			if (mode & KANA) {
+			if (mode & (KANA | JKANA)) {
 				if (!isjkana(s, i)) buf[j++] = s[i];
 				else {
 # ifdef	CODEEUC
 					buf[j++] = 0x8e;
 # endif
-					buf[j++] = s[i] | 0x80;
+					buf[j++] = jdecnv(s[i], io) | 0x80;
 				}
 			}
 			else if (mode & KANJI) {
@@ -615,35 +919,40 @@ int max, knj, asc;
 				else {
 # ifdef	CODEEUC
 					buf[j++] = s[i++] | 0x80;
-					buf[j++] = s[i] | 0x80;
+					buf[j++] = jdecnv(s[i], io) | 0x80;
 # else
 					u_char tmp[2];
 
 					tmp[0] = s[i++];
-					tmp[1] = s[i];
-					jis2sjis(&(buf[j]), tmp);
+					tmp[1] = jdecnv(s[i], io);
+					j2sj(&(buf[j]), tmp);
 					j += 2;
 # endif
 				}
 			}
+#  ifdef	CODEEUC
+			else if (iskna(s[i])) {
+				buf[j++] = 0x8e;
+				buf[j++] = s[i];
+			}
+#  endif
 			else buf[j++] = s[i];
 			break;
 	}
 	return(j);
 }
-#endif	/* !MSDOS && !_NOKANJICONV */
 
-#if	!MSDOS && !defined (_NOKANJIFCONV)
-static int NEAR tojis8(buf, s, max, knj, asc)
+# if	defined (FD) && (FD >= 2)
+static int NEAR tojis8(buf, s, max, knj, asc, io)
 char *buf;
 u_char *s;
-int max, knj, asc;
+int max, knj, asc, io;
 {
 	int i, j, mode;
 
 	mode = ASCII;
 	for (i = j = 0; s[i] && j < max - 7; i++, j++) {
-# ifdef	CODEEUC
+#  ifdef	CODEEUC
 		if (isekana(s, i)) {
 			if (mode == KANJI) {
 				buf[j++] = '\033';
@@ -654,7 +963,7 @@ int max, knj, asc;
 			buf[j] = s[++i];
 		}
 		else
-# endif
+#  endif
 		if (iskanji1(s, i)) {
 			if (mode != KANJI) {
 				buf[j++] = '\033';
@@ -662,12 +971,13 @@ int max, knj, asc;
 				buf[j++] = knj;
 			}
 			mode = KANJI;
-# ifdef	CODEEUC
+#  ifdef	CODEEUC
 			buf[j++] = s[i++] & ~0x80;
 			buf[j] = s[i] & ~0x80;
-# else
-			sjis2jis(&(buf[j++]), &(s[i++]));
-# endif
+#  else
+			sj2j(&(buf[j++]), &(s[i++]));
+#  endif
+			buf[j] = jcnv(buf[j], io);
 		}
 		else {
 			if (mode == KANJI) {
@@ -687,74 +997,21 @@ int max, knj, asc;
 	return(j);
 }
 
-static int NEAR fromjis8(buf, s, max, knj, asc)
+static int NEAR tojunet(buf, s, max, knj, asc, io)
 char *buf;
 u_char *s;
-int max, knj, asc;
-{
-	int i, j, mode;
-
-	mode = ASCII;
-	for (i = j = 0; s[i] && j < max - 1; i++)
-	switch (s[i]) {
-		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode = KANJI;
-				i += 2;
-				break;
-			}
-			else if (s[i + 1] == '(' && s[i + 2] == asc) {
-				mode = ASCII;
-				i += 2;
-				break;
-			}
-		default:
-			if (mode & KANJI) {
-				if (!isjis(s[i]) || !isjis(s[i + 1])) {
-					buf[j++] = s[i++];
-					buf[j++] = s[i];
-				}
-				else {
-# ifdef	CODEEUC
-					buf[j++] = s[i++] | 0x80;
-					buf[j++] = s[i] | 0x80;
-# else
-					u_char tmp[2];
-
-					tmp[0] = s[i++];
-					tmp[1] = s[i];
-					jis2sjis(&(buf[j]), tmp);
-					j += 2;
-# endif
-				}
-			}
-# ifdef	CODEEUC
-			else if (iskna(s[i])) {
-				buf[j++] = 0x8e;
-				buf[j++] = s[i];
-			}
-# endif
-			else buf[j++] = s[i];
-			break;
-	}
-	return(j);
-}
-
-static int NEAR tojunet(buf, s, max, knj, asc)
-char *buf;
-u_char *s;
-int max, knj, asc;
+int max, knj, asc, io;
 {
 	int i, j, mode;
 
 	mode = ASCII;
 	for (i = j = 0; s[i] && j < max - 7; i++, j++) {
-# ifdef	CODEEUC
+#  ifdef	CODEEUC
 		if (isekana(s, i)) {
 			i++;
-# else
+#  else
 		if (isskana(s, i)) {
-# endif
+#  endif
 			if (mode != JKANA) {
 				buf[j++] = '\033';
 				buf[j++] = '(';
@@ -762,6 +1019,7 @@ int max, knj, asc;
 			}
 			mode = JKANA;
 			buf[j] = s[i] & ~0x80;
+			buf[j] = jcnv(buf[j], io);
 		}
 		else if (iskanji1(s, i)) {
 			if (mode != KANJI) {
@@ -770,12 +1028,13 @@ int max, knj, asc;
 				buf[j++] = knj;
 			}
 			mode = KANJI;
-# ifdef	CODEEUC
+#  ifdef	CODEEUC
 			buf[j++] = s[i++] & ~0x80;
 			buf[j] = s[i] & ~0x80;
-# else
-			sjis2jis(&(buf[j++]), &(s[i++]));
-# endif
+#  else
+			sj2j(&(buf[j++]), &(s[i++]));
+#  endif
+			buf[j] = jcnv(buf[j], io);
 		}
 		else {
 			if (mode != ASCII) {
@@ -795,69 +1054,94 @@ int max, knj, asc;
 	return(j);
 }
 
-static int NEAR fromjunet(buf, s, max, knj, asc)
+static int NEAR toutf8(buf, s, max)
 char *buf;
 u_char *s;
-int max, knj, asc;
+int max;
 {
-	int i, j, mode;
+	u_short w;
+	int i, j;
 
-	mode = ASCII;
-	for (i = j = 0; s[i] && j < max - 1; i++)
-	switch (s[i]) {
-		case '\033':	/* ESC */
-			if (s[i + 1] == '$' && s[i + 2] == knj) {
-				mode = KANJI;
-				i += 2;
-				break;
-			}
-			else if (s[i + 1] == '(') {
-				if (s[i + 2] == 'I') {
-					mode = JKANA;
-					i += 2;
-					break;
-				}
-				else if (s[i + 2] == asc) {
-					mode = ASCII;
-					i += 2;
-					break;
-				}
-			}
-		default:
-			if (mode == JKANA) {
-				if (!isjkana(s, i)) buf[j++] = s[i];
-				else {
-# ifdef	CODEEUC
-					buf[j++] = 0x8e;
-# endif
-					buf[j++] = s[i] | 0x80;
-				}
-			}
-			else if (mode == KANJI) {
-				if (!isjis(s[i]) || !isjis(s[i + 1])) {
-					buf[j++] = s[i++];
-					buf[j++] = s[i];
-				}
-				else {
-# ifdef	CODEEUC
-					buf[j++] = s[i++] | 0x80;
-					buf[j++] = s[i] | 0x80;
-# else
-					u_char tmp[2];
+	for (i = j = 0; s[i] && j < max - 3; i++, j++) {
+#  ifdef	CODEEUC
+		if (isekana(s, i)) w = s[++i];
+		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
+			u_char tmp[2];
 
-					tmp[0] = s[i];
-					tmp[1] = s[i];
-					jis2sjis(&(buf[j]), tmp);
-					j += 2;
-# endif
-				}
-			}
-			else buf[j++] = s[i];
-			break;
+			j2sj(tmp, &(s[i++]));
+			w = (tmp[0] << 8) | tmp[1];
+		}
+#  else
+		if (isskana(s, i)) w = s[i];
+		else if (issjis1(s[i]) && issjis2(s[i + 1]))
+			w = (s[0] << 8) | s[1];
+#  endif
+		else w = s[i];
+		w = cnvunicode(w, 1);
+
+		if (w < 0x80) buf[j] = w;
+		else if (w < 0x800) {
+			buf[j] = 0xc0 | (w >> 6);
+			buf[j] = 0x80 | (w & 0x3f);
+		}
+		else {
+			buf[j++] = 0xe0 | (w >> 12);
+			buf[j++] = 0x80 | ((w >> 6) & 0x3f);
+			buf[j] = 0x80 | (w & 0x3f);
+		}
 	}
+	cnvunicode(0, -1);
 	return(j);
 }
 
+static int NEAR fromutf8(buf, s, max)
+char *buf;
+u_char *s;
+int max;
+{
+	u_short w;
+	int i, j, c1, c2;
+
+	for (i = j = 0; s[i] && j < max - 2; i++, j++) {
+		w = s[i];
+		if (w < 0x80);
+		else if ((w & 0xe0) == 0xc0 && (s[i + 1] & 0xc0) == 0x80)
+			w = ((w & 0x1f) << 6) | (s[++i] & 0x3f);
+		else {
+			w = ((w & 0x0f) << 6) | (s[++i] & 0x3f);
+			w = (w << 6) | (s[++i] & 0x3f);
+		}
+
+		w = cnvunicode(w, 0);
+		c1 = (w >> 8) & 0xff;
+		c2 = w & 0xff;
+		if (w > 0xa0 && w <= 0xdf) {
+#  ifdef	CODEEUC
+			buf[j++] = 0x8e;
+#  endif
+			buf[j] = w;
+		}
+		else if (issjis1(c1) && issjis2(c2)) {
+#  ifdef	CODEEUC
+			u_char tmp[2];
+
+			tmp[0] = c1;
+			tmp[1] = c2;
+			sj2j(&(buf[j]), tmp);
+			buf[j++] |= 0x80;
+			buf[j] |= 0x80;
+#  else
+			buf[j++] = c1;
+			buf[j] = c2;
+#  endif
+		}
+		else buf[j] = w;
+	}
+	cnvunicode(0, -1);
+	return(j);
+}
+
+#  ifndef	_NOKANJIFCONV
 static int NEAR bin2hex(buf, c)
 char *buf;
 int c;
@@ -879,22 +1163,22 @@ int max;
 	int i, j;
 
 	for (i = j = 0; s[i] && j < max - 2; i++, j++) {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 		if (isekana(s, i)) j += bin2hex(&(buf[j]), s[++i]) - 1;
-		else if (iseuc(s[i])) {
+		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
 			u_char tmp[2];
 
-			jis2sjis(tmp, &(s[i++]));
+			j2sj(tmp, &(s[i++]));
 			j += bin2hex(&(buf[j]), tmp[0]);
 			j += bin2hex(&(buf[j]), tmp[1]) - 1;
 		}
-# else
+#   else
 		if (isskana(s, i)) j += bin2hex(&(buf[j]), s[i]) - 1;
 		else if (issjis1(s[i]) && issjis2(s[i + 1])) {
 			j += bin2hex(&(buf[j]), s[i++]);
 			j += bin2hex(&(buf[j]), s[i]) - 1;
 		}
-# endif
+#   endif
 		else buf[j] = s[i];
 	}
 	return(j);
@@ -913,9 +1197,9 @@ int max;
 			c1 = tobin2(s, i + 1);
 			i += 2;
 			if (iskna(c1)) {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 				buf[j++] = 0x8e;
-# endif
+#   endif
 				buf[j] = c1;
 			}
 			else if (issjis1(c1)) {
@@ -930,18 +1214,18 @@ int max;
 					buf[j] = c2;
 				}
 				else {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 					u_char tmp[2];
 
 					tmp[0] = c1;
 					tmp[1] = c2;
-					sjis2jis(&(buf[j]), tmp);
+					sj2j(&(buf[j]), tmp);
 					buf[j++] |= 0x80;
 					buf[j] |= 0x80;
-# else
+#   else
 					buf[j++] = c1;
 					buf[j] = c2;
-# endif
+#   endif
 				}
 			}
 			else buf[j] = c1;
@@ -974,22 +1258,22 @@ int max;
 	int i, j;
 
 	for (i = j = 0; s[i] && j < max - 2; i++, j++) {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 		if (isekana(s, i)) j += bin2cap(&(buf[j]), s[++i]) - 1;
-		else if (iseuc(s[i])) {
+		else if (iseuc(s[i]) && iseuc(s[i + 1])) {
 			u_char tmp[2];
 
-			jis2sjis(tmp, &(s[i++]));
+			j2sj(tmp, &(s[i++]));
 			j += bin2cap(&(buf[j]), tmp[0]);
 			j += bin2cap(&(buf[j]), tmp[1]) - 1;
 		}
-# else
+#   else
 		if (isskana(s, i)) j += bin2cap(&(buf[j]), s[i]) - 1;
 		else if (issjis1(s[i]) && issjis2(s[i + 1])) {
 			j += bin2cap(&(buf[j]), s[i++]);
 			j += bin2cap(&(buf[j]), s[i]) - 1;
 		}
-# endif
+#   endif
 		else buf[j] = s[i];
 	}
 	return(j);
@@ -1008,9 +1292,9 @@ int max;
 			c1 = tobin2(s, i + 1);
 			i += 2;
 			if (iskna(c1)) {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 				buf[j++] = 0x8e;
-# endif
+#   endif
 				buf[j] = c1;
 			}
 			else if (issjis1(c1)) {
@@ -1025,18 +1309,18 @@ int max;
 					buf[j] = c2;
 				}
 				else {
-# ifdef	CODEEUC
+#   ifdef	CODEEUC
 					u_char tmp[2];
 
 					tmp[0] = c1;
 					tmp[1] = c2;
-					sjis2jis(&(buf[j]), tmp);
+					sj2j(&(buf[j]), tmp);
 					buf[j++] |= 0x80;
 					buf[j] |= 0x80;
-# else
+#   else
 					buf[j++] = c1;
 					buf[j] = c2;
-# endif
+#   endif
 				}
 			}
 			else buf[j] = c1;
@@ -1044,12 +1328,12 @@ int max;
 	}
 	return(j);
 }
-#endif	/* !MSDOS && !_NOKANJIFCONV */
+#  endif	/* !_NOKANJIFCONV */
+# endif	/* FD && (FD >= 2) */
 
-#if	!MSDOS && !defined (_NOKANJICONV)
-static char *NEAR _kanjiconv(buf, s, max, in, out, lenp)
+static char *NEAR _kanjiconv(buf, s, max, in, out, lenp, io)
 char *buf, *s;
-int max, in, out, *lenp;
+int max, in, out, *lenp, io;
 {
 	if (in == out || in == NOCNV || out == NOCNV) return(s);
 	switch (out) {
@@ -1074,37 +1358,30 @@ int max, in, out, *lenp;
 # endif
 					break;
 				case JIS7:
-					*lenp = fromjis7(buf, (u_char *)s,
-						max, 'B', 'B');
-					break;
-# if	!MSDOS && !defined (_NOKANJIFCONV)
+# if	defined (FD) && (FD >= 2)
 				case O_JIS7:
-					*lenp = fromjis7(buf, (u_char *)s,
-						max, '@', 'J');
-					break;
 				case JIS8:
-					*lenp = fromjis8(buf, (u_char *)s,
-						max, 'B', 'B');
-					break;
 				case O_JIS8:
-					*lenp = fromjis8(buf, (u_char *)s,
-						max, '@', 'J');
-					break;
 				case JUNET:
-					*lenp = fromjunet(buf, (u_char *)s,
-						max, 'B', 'B');
-					break;
 				case O_JUNET:
-					*lenp = fromjunet(buf, (u_char *)s,
-						max, '@', 'J');
+# endif	/* FD && (FD >= 2) */
+					*lenp = fromjis(buf, (u_char *)s,
+						max, io);
 					break;
+# if	defined (FD) && (FD >= 2)
+				case UTF8:
+					*lenp = fromutf8(buf, (u_char *)s,
+						max);
+					break;
+#  ifndef	_NOKANJIFCONV
 				case HEX:
 					*lenp = fromhex(buf, (u_char *)s, max);
 					break;
 				case CAP:
 					*lenp = fromcap(buf, (u_char *)s, max);
 					break;
-# endif	/* !MSDOS && !_NOKANJIFCONV */
+#  endif	/* !_NOKANJIFCONV */
+# endif	/* FD && (FD >= 2) */
 				default:
 					return(s);
 /*NOTREACHED*/
@@ -1115,31 +1392,36 @@ int max, in, out, *lenp;
 			*lenp = toenglish(buf, (u_char *)s, max);
 			break;
 		case JIS7:
-			*lenp = tojis7(buf, (u_char *)s, max, 'B', 'B');
+			*lenp = tojis7(buf, (u_char *)s, max, 'B', 'B', io);
 			break;
-# if	!MSDOS && !defined (_NOKANJIFCONV)
+# if	defined (FD) && (FD >= 2)
 		case O_JIS7:
-			*lenp = tojis7(buf, (u_char *)s, max, '@', 'J');
+			*lenp = tojis7(buf, (u_char *)s, max, '@', 'J', io);
 			break;
 		case JIS8:
-			*lenp = tojis8(buf, (u_char *)s, max, 'B', 'B');
+			*lenp = tojis8(buf, (u_char *)s, max, 'B', 'B', io);
 			break;
 		case O_JIS8:
-			*lenp = tojis8(buf, (u_char *)s, max, '@', 'J');
+			*lenp = tojis8(buf, (u_char *)s, max, '@', 'J', io);
 			break;
 		case JUNET:
-			*lenp = tojunet(buf, (u_char *)s, max, 'B', 'B');
+			*lenp = tojunet(buf, (u_char *)s, max, 'B', 'B', io);
 			break;
 		case O_JUNET:
-			*lenp = tojunet(buf, (u_char *)s, max, '@', 'J');
+			*lenp = tojunet(buf, (u_char *)s, max, '@', 'J', io);
 			break;
+		case UTF8:
+			*lenp = toutf8(buf, (u_char *)s, max);
+			break;
+#  ifndef	_NOKANJIFCONV
 		case HEX:
 			*lenp = tohex(buf, (u_char *)s, max);
 			break;
 		case CAP:
 			*lenp = tocap(buf, (u_char *)s, max);
 			break;
-# endif	/* !MSDOS && !_NOKANJIFCONV */
+#  endif	/* !_NOKANJIFCONV */
+# endif	/* FD && (FD >= 2) */
 		default:
 			return(s);
 /*NOTREACHED*/
@@ -1148,19 +1430,18 @@ int max, in, out, *lenp;
 	return(buf);
 }
 
-int kanjiconv(buf, s, max, in, out)
+int kanjiconv(buf, s, max, in, out, io)
 char *buf, *s;
-int max, in, out;
+int max, in, out, io;
 {
 	int len;
 
-	if (_kanjiconv(buf, s, max, in, out, &len) != buf)
+	if (_kanjiconv(buf, s, max, in, out, &len, io) != buf)
 		for (len = 0; s[len]; len++) buf[len] = s[len];
 	return(len);
 }
-#endif	/* !MSDOS && !_NOKANJICONV */
 
-#if	!MSDOS && !defined (_NOKANJIFCONV)
+# ifndef	_NOKANJIFCONV
 char *kanjiconv2(buf, s, max, in, out)
 char *buf, *s;
 int max, in, out;
@@ -1168,11 +1449,36 @@ int max, in, out;
 	char *cp;
 	int len;
 
-	if ((cp = _kanjiconv(buf, s, max, in, out, &len)) == buf)
+	if ((cp = _kanjiconv(buf, s, max, in, out, &len, L_FNAME)) == buf)
 		cp[len] = '\0';
 	return(cp);
 }
-#endif	/* !MSDOS && !_NOKANJIFCONV */
+# endif	/* !_NOKANJIFCONV */
+#endif	/* !MSDOS && !_NOKANJICONV */
+
+int kanjifputs(s, fp)
+char *s;
+FILE *fp;
+{
+#if	MSDOS || defined (_NOKANJICONV)
+	return(fputs(s, fp));
+#else
+	char *cp, *buf;
+	int len, max;
+
+	len = strlen(s);
+	max = len * 3 + 3;
+	buf = malloc2(max + 1);
+	cp = _kanjiconv(buf, s, max, DEFCODE, outputkcode, &len, L_OUTPUT);
+	if (cp == buf) {
+		buf[len] = '\0';
+		s = buf;
+	}
+	if (fputs(s, fp) < 0) len = -1;
+	free(buf);
+	return(len);
+#endif
+}
 
 int kanjiputs(s)
 char *s;
@@ -1181,13 +1487,14 @@ char *s;
 	cputs2(s);
 	return(strlen(s));
 #else
-	char *buf;
+	char *cp, *buf;
 	int len, max;
 
 	len = strlen(s);
-	max = len * 2 + 3;
+	max = len * 3 + 3;
 	buf = malloc2(max + 1);
-	if (_kanjiconv(buf, s, max, DEFCODE, outputkcode, &len) == buf) {
+	cp = _kanjiconv(buf, s, max, DEFCODE, outputkcode, &len, L_OUTPUT);
+	if (cp == buf) {
 		buf[len] = '\0';
 		s = buf;
 	}
@@ -1239,6 +1546,7 @@ int len, ptr;
 	char *dupl;
 
 	if (len >= 0) dupl = malloc2(len + 1);
+	else if (ptr < 0) dupl = malloc2(strlen(s) + 1);
 	else dupl = malloc2(strlen(&(s[ptr])) + 1);
 	strncpy3(dupl, s, &len, ptr);
 	kanjiputs(dupl);
