@@ -11,6 +11,12 @@
 #include "kctype.h"
 #include "kanji.h"
 
+#ifdef	_NOORIGSHELL
+#define	dosystem(s)		system(s)
+#else
+#include "system.h"
+#endif
+
 #ifndef	NOTZFILEH
 #include <tzfile.h>
 #endif
@@ -20,8 +26,6 @@
 extern int setcurdrv __P_((int, int));
 extern char *unixrealpath __P_((char *, char *));
 #else
-#include <pwd.h>
-#include <grp.h>
 #include <sys/file.h>
 #include <sys/param.h>
 #endif
@@ -33,10 +37,12 @@ extern int flushdrv __P_((int, VOID_T (*)__P_((VOID_A))));
 char Error[1024];
 #endif
 
+#ifdef	_NOORIGSHELL
 extern char **environ;
+#endif
 extern char fullpath[];
 extern char *origpath;
-extern char *findpattern;
+extern int hideclock;
 
 #ifndef	CHAR_BIT
 # ifdef	NBBY
@@ -51,8 +57,10 @@ extern char *findpattern;
 			| (((u_char *)(cp))[0] << (CHAR_BIT * 3)) )
 
 static char *NEAR _realpath2 __P_((char *, char *, int));
+#ifdef	_NOORIGSHELL
 static int NEAR _getenv2 __P_((char *, int, char **));
 static char **NEAR _putenv2 __P_((char *, char **));
+#endif
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
 static int NEAR tmcmp __P_((struct tm *, struct tm *));
@@ -61,20 +69,16 @@ static int NEAR tmcmp __P_((struct tm *, struct tm *));
 static long NEAR gettimezone __P_((struct tm *, time_t));
 #endif
 
-static char **environ2 = NULL;
+#ifdef	_NOORIGSHELL
+char **environ2 = NULL;
+#endif
+
 static char *lastpath = NULL;
 #ifndef	_NODOSDRIVE
 static char *unixpath = NULL;
 #endif
 
 
-int rename2(from, to)
-char *from, *to;
-{
-	if (!strpathcmp(from, to)) return(0);
-	return(Xrename(from, to));
-}
-
 int stat2(path, stp)
 char *path;
 struct stat *stp;
@@ -155,7 +159,8 @@ int rdlink;
 				buf[i] = '\0';
 				if (*buf == _SC_) strcpy(resolved, buf);
 				else {
-					*(cp - 1) = '\0';
+					if (cp - 1 > resolved) cp--;
+					*cp = '\0';
 					_realpath2(buf, resolved, rdlink);
 				}
 			}
@@ -172,10 +177,8 @@ int rdlink;
 	char tmp[MAXPATHLEN];
 	int drv;
 
-	if (path == resolved) {
-		strcpy(tmp, path);
-		path = tmp;
-	}
+	strcpy(tmp, path);
+	path = tmp;
 
 #if	MSDOS
 	drv = dospath("", NULL);
@@ -293,7 +296,7 @@ char *path;
 	return(0);
 }
 
-char *chdir3(path)
+int chdir3(path)
 char *path;
 {
 	char *cwd;
@@ -305,17 +308,23 @@ char *path;
 	if (!strcmp(path, ".")) cwd = NULL;
 	else if (!strcmp(path, "?")) path = origpath;
 	else if (!strcmp(path, "-")) {
-		if (!lastpath) return(".");
+		if (!lastpath) {
+			errno = ENOENT;
+			return(-1);
+		}
 		path = lastpath;
 	}
 #ifndef	_NODOSDRIVE
 	else if (!strcmp(path, "@")) {
-		if (!unixpath) return(".");
+		if (!unixpath) {
+			errno = ENOENT;
+			return(-1);
+		}
 		path = unixpath;
 	}
 	if ((drive = dospath3(fullpath))) flushdrv(drive, NULL);
 #endif
-	if (chdir2(path) < 0) return(NULL);
+	if (chdir2(path) < 0) return(-1);
 	if (!cwd) {
 		if (!Xgetwd(fullpath)) lostcwd(fullpath);
 	}
@@ -323,7 +332,7 @@ char *path;
 		if (findpattern) free(findpattern);
 		findpattern = NULL;
 	}
-	return(path);
+	return(0);
 }
 
 int mkdir2(path, mode)
@@ -469,9 +478,10 @@ int *lenp, ptr;
 			if (*lenp >= 0) (*lenp)++;
 			s1[i++] = s2[j++];
 		}
-		else
+#else
+		if (iskna(s2[j]));
 #endif
-		if (iskanji1(s2, j)) {
+		else if (iskanji1(s2, j)) {
 			if (*lenp >= 0 && i >= *lenp - 1) {
 				s1[i++] = ' ';
 				break;
@@ -484,6 +494,19 @@ int *lenp, ptr;
 			s1[i++] = ((s2[j++] + '@') & 0x7f);
 			continue;
 		}
+		else if (ismsb(s2[j])) {
+			int c;
+
+			c = s2[j++] & 0xff;
+			s1[i++] = '\\';
+			if (*lenp >= 0 && i >= *lenp) break;
+			s1[i++] = (c / (8 * 8)) + '0';
+			if (*lenp >= 0 && i >= *lenp) break;
+			s1[i++] = ((c % (8 * 8)) / 8) + '0';
+			if (*lenp >= 0 && i >= *lenp) break;
+			s1[i++] = (c % 8) + '0';
+			continue;
+		}
 		s1[i++] = s2[j++];
 	}
 
@@ -532,9 +555,16 @@ char *s;
 int atoi2(s)
 char *s;
 {
-	return((s && *s >= '0' && *s <= '9') ? atoi(s) : -1);
+	long n;
+
+	if (!s || !(s = evalnumeric(s, &n, 0)) || *s) return(-1);
+#if	MSDOS
+	if (n > MAXTYPE(int)) return(-1);
+#endif
+	return((int)n);
 }
 
+#ifdef	_NOORIGSHELL
 static int NEAR _getenv2(name, len, envp)
 char *name;
 int len;
@@ -569,9 +599,9 @@ char *s, **envp;
 	}
 	if (!cp) return(envp);
 
-#ifdef	DEBUG
-	_mtrace_file = "putenv";
-#endif
+# ifdef	DEBUG
+	_mtrace_file = "_putenv2";
+# endif
 	if (!envp) new = (char **)malloc((n + 2) * sizeof(char *));
 	else new = (char **)realloc(envp, (n + 2) * sizeof(char *));
 	if (!new) {
@@ -582,24 +612,30 @@ char *s, **envp;
 	new[n + 1] = (char *)NULL;
 	return(new);
 }
+#endif	/* _NOORIGSHELL */
 
 int putenv2(s)
 char *s;
 {
+#ifdef	_NOORIGSHELL
 	char **new;
-#if	MSDOS
+# if	MSDOS
 	char *cp;
 
 	for (cp = s; *cp && *cp != '='; cp++) *cp = toupper2(*cp);
-#endif	/* !MSDOS */
+# endif
 	if (!(new = _putenv2(s, environ))) return(-1);
 	environ = new;
 	return(0);
+#else	/* !_NOORIGSHELL */
+	return(putexportvar(s, -1));
+#endif	/* !_NOORIGSHELL */
 }
 
 char *getenv2(name)
 char *name;
 {
+#ifdef	_NOORIGSHELL
 	char **envpp[2];
 	int i, c, n, len;
 
@@ -615,35 +651,52 @@ char *name;
 		n = _getenv2(name + 3, len - 3, envpp[i]);
 		if (n >= 0 && envpp[i][n]) return(&(envpp[i][n][len - 3 + 1]));
 	}
+#else	/* !_NOORIGSHELL */
+	char *cp;
+
+	if ((cp = getshellvar(name, -1))) return(cp);
+	if (!strnpathcmp(name, "FD_", 3) && (cp = getshellvar(name + 3, -1)))
+		return(cp);
+#endif	/* !_NOORIGSHELL */
 	return(NULL);
 }
 
 int setenv2(name, value)
 char *name, *value;
 {
-	char *cp, **new;
+	char *cp;
 	int len;
-#if	MSDOS
+#ifdef	_NOORIGSHELL
+	char **new;
+# if	MSDOS
 	int i;
+# endif
 #endif
 
+	len = strlen(name);
 	if (!value) cp = name;
 	else {
-		len = strlen(name);
 		cp = malloc2(len + strlen(value) + 2);
 		memcpy(cp, name, len);
-#if	MSDOS
+#if	MSDOS && defined (_NOORIGSHELL)
 		for (i = 0; i < len ; i++) cp[i] = toupper2(cp[i]);
 #endif
-		cp[len++] = '=';
-		strcpy(&(cp[len]), value);
+		cp[len] = '=';
+		strcpy(&(cp[len + 1]), value);
 	}
+#ifdef	_NOORIGSHELL
 	if (!(new = _putenv2(cp, environ2))) return(-1);
 	environ2 = new;
+#else	/* !_NOORIGSHELL */
+	if (putshellvar(cp, len) < 0) {
+		free(cp);
+		return(-1);
+	}
+#endif	/* !_NOORIGSHELL */
 	return(0);
 }
 
-#ifdef	DEBUG
+#if	defined (DEBUG) && defined (_NOORIGSHELL)
 VOID freeenv(VOID_A)
 {
 	int i;
@@ -657,62 +710,41 @@ VOID freeenv(VOID_A)
 		free(environ2);
 	}
 }
-#endif
-
-/*ARGSUSED*/
-int printenv(argc, argv, comline)
-int argc;
-char *argv[];
-int comline;
-{
-	int i, n;
-
-	if (argc >= 3) return(-1);
-	if (!comline) return(0);
-	if (argc == 2) {
-		n = _getenv2(argv[1], strlen(argv[1]), environ2);
-		if (n >= 0 && environ2[n]) cprintf2("%s\r\n", environ2[n]);
-		return(1);
-	}
-	if (!environ2) return(0);
-	for (i = n = 0; environ2[i]; i++) {
-		cprintf2("%s\r\n", environ2[i]);
-		if (++n >= n_line - 1) {
-			n = 0;
-			warning(0, HITKY_K);
-		}
-	}
-	return(n);
-}
+#endif	/* DEBUG && _NOORIGSHELL */
 
 int system2(command, noconf)
 char *command;
 int noconf;
 {
-	int ret;
+	int n, ret;
 
 	if (!command || !*command) return(0);
+	n = sigvecset(0);
 	if (noconf >= 0) {
 		locate(0, n_line - 1);
 		putterm(l_clear);
-		tflush();
-		if (noconf) putterms(t_end);
-		putterms(t_nokeypad);
-		tflush();
+		if (n && noconf) putterms(t_end);
 	}
-	cooked2();
-	echo2();
-	nl2();
-	tabs();
-	ret = system(command);
-	raw2();
-	noecho2();
-	nonl2();
-	notabs();
-	if (ret > 127 || !noconf) warning(0, HITKY_K);
-	if (noconf >= 0) {
-		if (noconf) putterms(t_init);
-		putterms(t_keypad);
+	stdiomode();
+	ret = dosystem(command);
+	if (noconf > 0) {
+		if (ret >= 127) {
+			fputs("\007\n", stderr);
+			kanjifputs(HITKY_K, stderr);
+			fflush(stderr);
+			ttyiomode();
+			keyflush();
+			getkey2(0);
+			stdiomode();
+			fputc('\n', stderr);
+		}
+		if (n) putterms(t_init);
+	}
+	ttyiomode();
+	sigvecset(n);
+	if (!noconf || (noconf < 0 && ret >= 127)) {
+		hideclock = 1;
+		warning(0, HITKY_K);
 	}
 	return(ret);
 }
@@ -725,66 +757,6 @@ char *getwd2(VOID_A)
 	return(strdup2(cwd));
 }
 
-#if	!MSDOS
-char *getpwuid2(uid)
-uid_t uid;
-{
-	static strtable *uidlist = NULL;
-	static int maxuid = 0;
-	struct passwd *pwd;
-	int i;
-
-	for (i = 0; i < maxuid; i++)
-		if (uid == uidlist[i].no) return(uidlist[i].str);
-
-#ifdef	DEBUG
-	_mtrace_file = "getpwuid";
-#endif
-	if ((pwd = getpwuid(uid))) {
-#ifdef	DEBUG
-		_mtrace_file = "getpwuid2(1)";
-#endif
-		uidlist = b_realloc(uidlist, maxuid, strtable);
-		uidlist[maxuid].no = pwd -> pw_uid;
-#ifdef	DEBUG
-		_mtrace_file = "getpwuid2(2)";
-#endif
-		uidlist[maxuid].str = strdup2(pwd -> pw_name);
-		return(uidlist[maxuid++].str);
-	}
-	return(NULL);
-}
-
-char *getgrgid2(gid)
-gid_t gid;
-{
-	static strtable *gidlist = NULL;
-	static int maxgid = 0;
-	struct group *grp;
-	int i;
-
-	for (i = 0; i < maxgid; i++)
-		if (gid == gidlist[i].no) return(gidlist[i].str);
-
-#ifdef	DEBUG
-	_mtrace_file = "getgrgid";
-#endif
-	if ((grp = getgrgid(gid))) {
-#ifdef	DEBUG
-		_mtrace_file = "getgrgid2(1)";
-#endif
-		gidlist = b_realloc(gidlist, maxgid, strtable);
-		gidlist[maxgid].no = grp -> gr_gid;
-#ifdef	DEBUG
-		_mtrace_file = "getgrgid2(2)";
-#endif
-		gidlist[maxgid].str = strdup2(grp -> gr_name);
-		return(gidlist[maxgid++].str);
-	}
-	return(NULL);
-}
-#endif
-
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
 static int NEAR tmcmp(tm1, tm2)
@@ -965,8 +937,9 @@ struct tm *tm;
 #endif
 }
 
-char *fgets2(fp)
+char *fgets2(fp, nulcnv)
 FILE *fp;
+int nulcnv;
 {
 	char *cp;
 	long i, size;
@@ -982,8 +955,12 @@ FILE *fp;
 			break;
 		}
 		cp = c_realloc(cp, i, size);
-		cp[i] = (c) ? c : '\n';
+		if (nulcnv && !c) c = '\n';
+		cp[i] = c;
 	}
+#if	MSDOS
+	if (i > 0 && cp[i - 1] == '\r') i--;
+#endif
 	cp[i++] = '\0';
 	return(realloc2(cp, i));
 }
