@@ -27,18 +27,18 @@
 #if	MSDOS
 #include <process.h>
 #include "unixemu.h"
-#ifndef	FD
-#include <dos.h>
-#include <io.h>
-#define	DS_IRDONLY	001
-#define	DS_IHIDDEN	002
-#define	DS_IFSYSTEM	004
-#define	DS_IFLABEL	010
-#define	DS_IFDIR	020
-#define	DS_IARCHIVE	040
-#define	SEARCHATTRS	(DS_IRDONLY | DS_IHIDDEN | DS_IFSYSTEM \
+# ifndef	FD
+# include <dos.h>
+# include <io.h>
+# define	DS_IRDONLY	001
+# define	DS_IHIDDEN	002
+# define	DS_IFSYSTEM	004
+# define	DS_IFLABEL	010
+# define	DS_IFDIR	020
+# define	DS_IARCHIVE	040
+# define	SEARCHATTRS	(DS_IRDONLY | DS_IHIDDEN | DS_IFSYSTEM \
 			| DS_IFDIR | DS_IARCHIVE)
-#endif
+# endif	/* !FD */
 #define	EXTWIDTH	4
 # if defined (DJGPP) && DJGPP < 2
 # include <dir.h>
@@ -157,6 +157,7 @@ static int NEAR _evalwild __P_((int, char ***, char *, char *, int,
 static int NEAR calchash __P_((char *));
 static VOID NEAR inithash __P_((VOID_A));
 static hashlist *NEAR newhash __P_((char *, char *, int, hashlist *));
+static VOID NEAR freehash __P_((hashlist **));
 static hashlist *NEAR findhash __P_((char *, int));
 static VOID NEAR rmhash __P_((char *, int));
 #endif
@@ -223,15 +224,13 @@ static int NEAR evalhome __P_((char **, int, char **));
 			(((n) + 1 < (size)) \
 			? (ptr) : realloc2(ptr, (size) *= 2))
 
-static int skipdotfile = 0;
-static char *wildsymbol1 = "?";
-static char *wildsymbol2 = "*";
 #ifndef	_NOUSEHASH
 hashlist **hashtable = NULL;
 #endif
 char *(*getvarfunc)__P_((char *, int)) = NULL;
 int (*putvarfunc)__P_((char *, int)) = NULL;
 int (*getretvalfunc)__P_((VOID_A)) = NULL;
+long (*getpidfunc)__P_((VOID_A)) = NULL;
 long (*getlastpidfunc)__P_((VOID_A)) = NULL;
 char **(*getarglistfunc)__P_((VOID_A)) = NULL;
 char *(*getflagfunc)__P_((VOID_A)) = NULL;
@@ -240,10 +239,6 @@ VOID (*exitfunc)__P_((VOID_A)) = NULL;
 char *(*backquotefunc)__P_((char *)) = NULL;
 #if	!MSDOS
 int pathignorecase = 0;
-static uidtable *uidlist = NULL;
-static int maxuid = 0;
-static gidtable *gidlist = NULL;
-static int maxgid = 0;
 #endif	/* !MSDOS */
 #ifndef	LSI_C
 u_char uppercase[256] = {
@@ -354,6 +349,16 @@ u_char kctypetable[256] = {
 # endif	/* FD || FDSH */
 #endif	/* !LSI_C */
 
+static int skipdotfile = 0;
+static char *wildsymbol1 = "?";
+static char *wildsymbol2 = "*";
+#if	!MSDOS
+static uidtable *uidlist = NULL;
+static int maxuid = 0;
+static gidtable *gidlist = NULL;
+static int maxgid = 0;
+#endif	/* !MSDOS */
+
 
 #ifndef	FD
 static VOID NEAR allocerror(VOID_A)
@@ -850,7 +855,7 @@ int len;
 # if	MSDOS
 	re_ignore_case = 1;
 # endif
-	skipdotfile = (*s == '*' || *s == '?');
+	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
 	s = cnvregexp(s, len);
 	re_comp(s);
 	free(s);
@@ -883,7 +888,7 @@ int len;
 {
 	reg_t *re;
 
-	skipdotfile = (*s == '*' || *s == '?');
+	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
 	s = cnvregexp(s, len);
 	re = (reg_t *)malloc2(sizeof(reg_t));
 # if	MSDOS
@@ -927,7 +932,7 @@ int len;
 {
 	reg_t *re;
 
-	skipdotfile = (*s == '*' || *s == '?');
+	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
 	s = cnvregexp(s, len);
 	re = regcmp(s, 0);
 	free(s);
@@ -956,9 +961,9 @@ int len;
 {
 	reg_t *re;
 	char *cp, *paren;
-	int i, j, n, plen, size, meta, quote;
+	int i, j, n, plen, size, metachar, quote;
 
-	skipdotfile = (*s == '*' || *s == '?');
+	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
 	if (len < 0) len = strlen(s);
 	paren = NULL;
 	n = plen = size = 0;
@@ -966,14 +971,14 @@ int len;
 	re[0] = NULL;
 	for (i = 0, quote = '\0'; i < len; i++) {
 		cp = NULL;
-		meta = 0;
+		metachar = 0;
 		if (s[i] == quote) {
 			quote = '\0';
 			continue;
 		}
 		else if (quote == '\'');
 		else if (isnmeta(s, i, quote, len)) {
-			meta = 1;
+			metachar = 1;
 			i++;
 		}
 		else if (quote);
@@ -985,9 +990,9 @@ int len;
 		if (paren) {
 			paren = c_realloc(paren, plen + 1, size);
 
-			if (quote || meta) {
+			if (quote || metachar) {
 #ifdef	BASHSTYLE
-	/* bash treats a meta character in "[]" as a character itself */
+	/* bash treats a character quoted by \ in "[]" as a character itself */
 				paren[plen++] = PMETA;
 # if	!MSDOS
 				if (!pathignorecase) paren[plen++] = s[i];
@@ -1074,6 +1079,8 @@ int len;
 				paren = c_malloc(size);
 				plen = 0;
 				break;
+			default:
+				break;
 		}
 
 		if (paren) continue;
@@ -1142,11 +1149,11 @@ char *s;
 		c2 = '\0';
 		for (; re[n1][i]; i++) {
 #ifdef	BASHSTYLE
-	/* bash treats a meta character in "[]" as a character itself */
+	/* bash treats a character quoted by \ in "[]" as a character itself */
 			if (re[n1][i] == PMETA && re[n1][i + 1]) i++;
-			if (re[n1][i] == '-' && re[n1][i + 1] && c2)
+			else if (re[n1][i] == '-' && re[n1][i + 1] && c2)
 #else
-			if (re[n1][i] == '-')
+			if (re[n1][i] == '-' && c2)
 #endif
 			{
 				beg = c2;
@@ -1155,7 +1162,7 @@ char *s;
 			c2 = (u_char)(re[n1][i]);
 #ifdef	CODEEUC
 			if (isekana(re[n1], i))
-				c1 = (c1 << 8) + (u_char)(re[n1][++i]);
+				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
 			else
 #endif
 			if (iskanji1(re[n1], i))
@@ -1203,36 +1210,35 @@ reg_t *re;
 # endif		/* !USEREGCOMP */
 #endif		/* !USERE_COMP */
 
-static char *NEAR catpath(path, file, plenp, flen, overwrite)
+static char *NEAR catpath(path, file, plenp, flen, isoverwrite)
 char *path, *file;
-int *plenp, flen, overwrite;
+int *plenp, flen, isoverwrite;
 {
 	char *new;
-	int i;
+	int i, sc;
 
-	if (!*plenp) {
-		new = malloc2(flen + 1);
-		strncpy2(new, file, flen);
-	}
-	else {
-#if	MSDOS
-		if (_dospath(path)) i = 2;
-		else
+#if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
+	if (*plenp >= 2 && _dospath(path)) i = 2;
+	else
 #endif
-		i = 0;
+	i = 0;
 
+	if (*plenp <= i) sc = 0;
+	else {
+		sc = 1;
 		if (path[i] == _SC_) {
-			for (i++; path[i] == _SC_; i++);
-			if (!path[i]) *plenp = i - 1;
+			for (i++; i < *plenp; i++) if (path[i] != _SC_) break;
+			if (i >= *plenp) sc = 0;
 		}
-		if (overwrite) new = realloc2(path, *plenp + 1 + flen + 1);
-		else {
-			new = malloc2(*plenp + 1 + flen + 1);
-			strncpy(new, path, *plenp);
-		}
-		new[(*plenp)++] = _SC_;
-		strncpy2(new + *plenp, file, flen);
 	}
+
+	if (isoverwrite) new = realloc2(path, *plenp + sc + flen + 1);
+	else {
+		new = malloc2(*plenp + sc + flen + 1);
+		strncpy(new, path, *plenp);
+	}
+	if (sc) new[(*plenp)++] = _SC_;
+	strncpy2(new + *plenp, file, flen);
 	*plenp += flen;
 	return(new);
 }
@@ -1510,7 +1516,8 @@ hashlist *next;
 	return(new);
 }
 
-VOID freehash(hashlist **htable)
+static VOID NEAR freehash(htable)
+hashlist **htable;
 {
 	hashlist *hp, *next;
 	int i;
@@ -1639,6 +1646,13 @@ char *com, *search;
 	static char *searchcwd = NULL;
 	char buf[MAXPATHLEN];
 	int n;
+
+	if (!com && !search) {
+		freehash(hpp);
+		if (searchcwd) free(searchcwd);
+		searchcwd = NULL;
+		return(CM_NOTFOUND);
+	}
 #endif
 
 #if	MSDOS
@@ -2093,7 +2107,9 @@ int c, quoted;
 			cp = tmp;
 			break;
 		case '$':
-			sprintf(tmp, "%ld", (long)getpid());
+			if (!getpidfunc) pid = getpid();
+			else pid = (*getpidfunc)();
+			sprintf(tmp, "%ld", pid);
 			cp = tmp;
 			break;
 		case '!':
@@ -2173,7 +2189,7 @@ char *arg;
 int olen, nlen;
 {
 	if (nlen <= olen) return(buf);
-	buf = realloc2(buf, ptr + nlen + strlen(arg) - olen + 1);
+	buf = realloc2(buf, ptr + nlen + (int)strlen(arg) - olen + 1);
 	return(buf);
 }
 
@@ -2225,6 +2241,7 @@ int quoted;
 			else if (quote == '\'');
 			else if (ismeta(top, i, quote)) {
 				if (len < 0) return(-1);
+				i++;
 			}
 			else if (quote);
 			else if (top[i] == '\'' || top[i] == '"') {
@@ -2573,9 +2590,12 @@ int stripq, backq;
 			}
 		}
 		else if (ismeta(cp, 0, quote)) {
-			if (j >= 0);
-			else buf[i++] = *cp;
 			cp++;
+			if (*cp == '$');
+			else if (backq && *cp == '`');
+			else if (stripq && (*cp == '\'' || *cp == '"'));
+			else if (j >= 0) bbuf[j++] = PMETA;
+			else buf[i++] = PMETA;
 		}
 		else if (backq && backquotefunc && *cp == '`') {
 			if (j < 0) j = 0;
@@ -2618,10 +2638,9 @@ int stripq, backq;
 	return(buf);
 }
 
-int evalifs(argc, argvp, ifs, iscom)
+int evalifs(argc, argvp, ifs)
 int argc;
 char ***argvp, *ifs;
-int iscom;
 {
 	char *cp;
 	int i, j, n, quote;
@@ -2663,7 +2682,7 @@ int iscom;
 				break;
 			}
 		}
-		if (iscom && !i) {
+		if (!i) {
 			free((*argvp)[n]);
 			for (i = n; i < argc; i++)
 				(*argvp)[i] = (*argvp)[i + 1];
@@ -2674,10 +2693,10 @@ int iscom;
 	return(argc);
 }
 
-int evalglob(argc, argvp, stripq, iscom)
+int evalglob(argc, argvp, stripq)
 int argc;
 char ***argvp;
-int stripq, iscom;
+int stripq;
 {
 	char *cp, **wild;
 	int i, j, n, w, size, quote;
@@ -2709,7 +2728,6 @@ int stripq, iscom;
 				quote = (*argvp)[n][i];
 				if (stripq) continue;
 			}
-			else if (iscom && !n);
 			else if (!strchr("?*[", (*argvp)[n][i]));
 			else if (wild = evalwild((*argvp)[n])) {
 				for (w = 0; wild[w]; w++);
@@ -2733,7 +2751,7 @@ int stripq, iscom;
 				for (w = 0; wild[w]; w++)
 					(*argvp)[n + w] = wild[w];
 				free(wild);
-				n += w;
+				n += w - 1;
 				break;
 			}
 
