@@ -33,17 +33,22 @@ typedef struct _redirectlist {
 	u_char type;
 	int new;
 	int old;
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	char *fakepipe;
+	int dosfd;
+#endif
 	struct _redirectlist *next;
 } redirectlist;
 
-#define	MD_NORMAL	000
-#define	MD_READ		001
-#define	MD_WRITE	002
-#define	MD_RDWR		003
-#define	MD_APPEND	004
-#define	MD_FILEDESC	010
-#define	MD_WITHERR	020
-#define	MD_HEREDOC	040
+#define	MD_NORMAL	0000
+#define	MD_READ		0001
+#define	MD_WRITE	0002
+#define	MD_RDWR		0003
+#define	MD_APPEND	0004
+#define	MD_FILEDESC	0010
+#define	MD_WITHERR	0020
+#define	MD_HEREDOC	0040
+#define	MD_FORCED	0100
 
 typedef struct _command_t {
 	hashlist *hash;
@@ -119,6 +124,7 @@ typedef struct _syntaxtree {
 #define	OP_OR	4
 #define	OP_PIPE	5
 #define	OP_NOT	6
+#define	OP_NOWN	7
 
 typedef struct _shbuiltintable {
 	int (NEAR *func)__P_((syntaxtree *));
@@ -128,6 +134,7 @@ typedef struct _shbuiltintable {
 
 #define	BT_NOGLOB	0001
 #define	BT_RESTRICT	0002
+#define	BT_POSIXSPECIAL	0004
 
 #define	SMPREV	4
 typedef struct _statementtable {
@@ -216,7 +223,7 @@ extern char *ifs_set;
 extern char **shellvar;
 extern char **envvar;
 
-extern VOID prepareexit __P_((VOID_A));
+extern VOID prepareexit __P_((int));
 #ifndef	NOJOB
 extern VOID killjob __P_((VOID_A));
 #endif
