@@ -100,10 +100,15 @@ static int NEAR putdostime __P_((u_short *, u_short *, time_t));
 
 #ifndef	_NODOSDRIVE
 int dosdrive = 0;
+#endif
+
+#ifndef	_NODOSDRIVE
 static drvinfo drvlist['Z' - 'A' + 1];
 static int maxdrive = -1;
 #endif
-
+static int dos7access = 0;
+#define	D7_DOSVER	017
+#define	D7_CAPITAL	020
 static u_short doserrlist[] = {
 	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 18, 65, 80
 };
@@ -148,9 +153,7 @@ static u_char int25bin[] = {
 };
 static int (far *doint25)__P_((VOID_A)) = NULL;
 #endif
-static int dos7access = 0;
-#define	D7_DOSVER	017
-#define	D7_CAPITAL	020
+
 #ifndef	LSI_C
 # ifdef	FD
 extern u_char uppercase[256];
@@ -520,8 +523,24 @@ int drive, nodir;
 #ifndef	_NOUSELFN
 int getdosver(VOID_A)
 {
-	if (!(dos7access & D7_DOSVER))
-		dos7access |= ((u_char)bdos(0x30, 0, 0) & D7_DOSVER);
+	struct SREGS sreg;
+	__dpmi_regs reg;
+	int ver;
+
+	if (!(dos7access & D7_DOSVER)) {
+		reg.x.ax = 0x3000;
+		int21call(&reg, &sreg);
+		ver = reg.h.al;
+		if (reg.x.ax == 0x0005) {
+			reg.x.ax = 0x3306;
+			int21call(&reg, &sreg);
+			if (reg.h.al != 0xff && reg.x.bx == 0x3205) {
+				/* Windows NT/2000/XP command prompt */
+				ver = 7;
+			}
+		}
+		dos7access |= (ver & D7_DOSVER);
+	}
 	return(dos7access & D7_DOSVER);
 }
 
@@ -2231,7 +2250,7 @@ struct stat *stp;
 
 #ifdef	_NOUSELFN
 	if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-		if (errno) return(-1);
+		if (errno && errno != ENOENT) return(-1);
 		if (!strcmp(path, ".."))
 			i = dos_findfirst(".", SEARCHATTRS, &dbuf);
 		else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
@@ -2250,7 +2269,7 @@ struct stat *stp;
 	if (i <= 0) {
 		fd = -1;
 		if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-			if (errno) return(-1);
+			if (errno && errno != ENOENT) return(-1);
 			if (!strcmp(path, ".."))
 				i = dos_findfirst(".", SEARCHATTRS, &dbuf);
 			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
